/*
===========================================================================

Doom 3 GPL Source Code
Copyright (C) 1999-2011 id Software LLC, a ZeniMax Media company.

This file is part of the Doom 3 GPL Source Code (?Doom 3 Source Code?).

Doom 3 Source Code is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Doom 3 Source Code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Doom 3 Source Code.  If not, see <http://www.gnu.org/licenses/>.

In addition, the Doom 3 Source Code is also subject to certain additional terms. You should have received a copy of these additional terms immediately following the terms and conditions of the GNU General Public License which accompanied the Doom 3 Source Code.  If not, please request a copy in writing from id Software at the address below.

If you have questions concerning this license or the applicable additional terms, you may contact in writing id Software LLC, c/o ZeniMax Media Inc., Suite 120, Rockville, Maryland 20850 USA.

===========================================================================
*/


#include "../precompiled.h"
#pragma hdrstop

#include "Simd_Generic.h"
#include "Simd_Neon.h"
#include <math.h>
#include <float.h>

#ifdef __ARM_NEON__
#include <arm_neon.h>
#endif

// Doom3 SIMD Library version 0.5
// Patrick Flanagan (pflanagan@apple.com)
// Sanjay Patel (spatel@apple.com)
// Architecture & Performance Group, Apple Computer


//===============================================================
//
//	AltiVec implementation of idSIMDProcessor
//
//===============================================================

#if defined(__ARM_NEON__)

#define NOT_4BYTE_ALIGNED( x ) ( ( (unsigned long)&x & 0x03) != 0 )


float32x4_t SIMD_SP_tiny = {1e-10f, 1e-10f, 1e-10f, 1e-10f};
/*
============
idSIMD_NEON::GetName
============
*/
const char *idSIMD_NEON::GetName(void) const
{
	return "NEON";
}

/*
============
idSIMD_NEON::Add

  dst[i] = constant + src[i];
============
*/
void VPCALL idSIMD_NEON::Add(float *dst, const float constant, const float *src, const int count)
{
float32x2_t constx2 = vdup_n_f32(constant);
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vadd_f32(vld1_f32(src+_IX),  constx2));
if (_IX < count) dst[(_IX)] = src[(_IX)] + constant;
}

/*
============
idSIMD_NEON::Add

  dst[i] = src0[i] + src1[i];
============
*/
void VPCALL idSIMD_NEON::Add(float *dst, const float *src0, const float *src1, const int count)
{
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vadd_f32(vld1_f32(src0+_IX),  vld1_f32(src1+_IX)));
if (_IX < count) dst[_IX] = src0[_IX] + src1[_IX];
}


/*
============
idSIMD_NEON::Sub

  dst[i] = constant - src[i];
============
*/
void VPCALL idSIMD_NEON::Sub(float *dst, const float constant, const float *src, const int count)
{
float32x2_t constx2 = vdup_n_f32(constant);
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vsub_f32(constx2, vld1_f32(src+_IX)));
if (_IX < count) dst[_IX] = constant - src[_IX];
}

/*
============
idSIMD_NEON::Sub

  dst[i] = src0[i] - src1[i];
============
*/
void VPCALL idSIMD_NEON::Sub(float *dst, const float *src0, const float *src1, const int count)
{
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vsub_f32(vld1_f32(src0+_IX),  vld1_f32(src1+_IX)));
if (_IX < count) dst[(_IX)] = src0[_IX] - src1[_IX];
}

/*
============
idSIMD_NEON::Mul

  dst[i] = constant * src[i];
============
*/
void VPCALL idSIMD_NEON::Mul(float *dst, const float constant, const float *src0, const int count)
{
float32x2_t constx2 = vdup_n_f32(constant);
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmul_f32(constx2, vld1_f32(src0+_IX)));
if (_IX < count) dst[_IX] = constant * src0[_IX];
}

/*
============
idSIMD_NEON::Mul

  dst[i] = src0[i] * src1[i];
============
*/
void VPCALL idSIMD_NEON::Mul(float *dst, const float *src0, const float *src1, const int count)
{
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmul_f32(vld1_f32(src0+_IX), vld1_f32(src1+_IX)));
if (_IX < count) dst[_IX] = src0[_IX] * src1[_IX];
}

/*
============
idSIMD_NEON::Div

  dst[i] = constant / divisor[i];
============
*/
/*void VPCALL idSIMD_NEON::Div(float *dst, const float constant, const float *divisor, const int count)
{
	float c = constant;
#define OPER(X) (dst[(X)] = (c / divisor[(X)]))
	UNROLL4(OPER)
#undef OPER
}*/

/*
============
idSIMD_NEON::MulAdd

  dst[i] += constant * src[i];
============
*/
void VPCALL idSIMD_NEON::MulAdd(float *dst, const float constant, const float *src, const int count)
{
float32x2_t constx2 = vdup_n_f32(constant);
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmla_f32(vld1_f32(dst+_IX), constx2, vld1_f32(src+_IX)));
if (_IX < count) dst[_IX] += constant * src[_IX];
}

/*
============
idSIMD_NEON::MulAdd

  dst[i] += src0[i] * src1[i];
============
*/
void VPCALL idSIMD_NEON::MulAdd(float *dst, const float *src0, const float *src1, const int count)
{
int _IX, _NM = count&0xfffffffe;
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmla_f32(vld1_f32(dst+_IX), vld1_f32(src0+_IX), vld1_f32(src1+_IX)));
if (_IX < count) dst[_IX] += src0[_IX] * src1[_IX];
}

/*
============
idSIMD_NEON::MulSub

  dst[i] -= constant * src[i];
============
*/
void VPCALL idSIMD_NEON::MulSub(float *dst, const float constant, const float *src, const int count)
{
float32x2_t constx2 = vdup_n_f32(constant);
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmls_f32(vld1_f32(dst+_IX), constx2, vld1_f32(src+_IX)));
if (_IX < count) dst[_IX] -= constant * src[_IX];
}

/*
============
idSIMD_NEON::MulSub

  dst[i] -= src0[i] * src1[i];
============
*/
void VPCALL idSIMD_NEON::MulSub(float *dst, const float *src0, const float *src1, const int count)
{
int _IX, _NM = count&0xfffffffe;
for (_IX=0;_IX<_NM;_IX+=2)
	vst1_f32(dst+_IX, vmls_f32(vld1_f32(dst+_IX), vld1_f32(src0+_IX), vld1_f32(src1+_IX)));
if (_IX < count) dst[_IX] -= src0[_IX] * src1[_IX];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = constant * src[i];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idVec3 &constant, const idVec3 *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t xx=vdup_n_f32(constant.x), yy=vdup_n_f32(constant.y), zz=vdup_n_f32(constant.z);
for (_IX=0;_IX<_NM;_IX+=2) {
	float32x2x3_t src2=vld3_f32((float32_t*)(src+_IX));
	vst1_f32(dst+_IX, src2.val[0]*xx+src2.val[1]*yy+src2.val[2]*zz);
}
if (_IX < count) dst[_IX] = constant * src[_IX];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = constant * src[i].Normal() + src[i][3];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idVec3 &constant, const idPlane *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t xx=vdup_n_f32(constant[0]), yy=vdup_n_f32(constant[1]), zz=vdup_n_f32(constant[2]);
for (_IX=0;_IX<_NM;_IX+=2) {
	float32x2x4_t src2=vld4_f32((float32_t*)(src+_IX));
	vst1_f32(dst+_IX, src2.val[0]*xx+src2.val[1]*yy+src2.val[2]*zz+src2.val[3]);
}
if (_IX < count) dst[_IX] = constant * src[_IX].Normal() + src[_IX][3];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = constant * src[i].xyz;
============
*/
/*void VPCALL idSIMD_NEON::Dot(float *dst, const idVec3 &constant, const idDrawVert *src, const int count)
{
#define OPER(X) dst[(X)] = constant * src[(X)].xyz;
	UNROLL1(OPER)
#undef OPER
}
*/

/*
============
idSIMD_NEON::Dot

  dst[i] = constant.Normal() * src[i] + constant[3];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idPlane &constant, const idVec3 *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t xx=vdup_n_f32(constant[0]), yy=vdup_n_f32(constant[1]), zz=vdup_n_f32(constant[2]), ww=vdup_n_f32(constant[3]);
for (_IX=0;_IX<_NM;_IX+=2) {
	float32x2x3_t src2=vld3_f32((float32_t*)(src+_IX));
	vst1_f32(dst+_IX, src2.val[0]*xx+src2.val[1]*yy+src2.val[2]*zz+ww);
}
if (_IX < count) dst[_IX] = constant.Normal() * src[_IX] + constant[3];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = constant.Normal() * src[i].Normal() + constant[3] * src[i][3];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idPlane &constant, const idPlane *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t xx=vdup_n_f32(constant[0]), yy=vdup_n_f32(constant[1]), zz=vdup_n_f32(constant[2]), ww=vdup_n_f32(constant[3]);
for (_IX=0;_IX<_NM;_IX+=2) {
	float32x2x4_t src2=vld4_f32((float32_t*)(src+_IX));
	vst1_f32(dst+_IX, src2.val[0]*xx+src2.val[1]*yy+src2.val[2]*zz+src2.val[3]*ww);
}
if (_IX < count) dst[_IX] = constant.Normal() * src[_IX].Normal() + constant[3]*src[_IX][3];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = constant.Normal() * src[i].xyz + constant[3];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idPlane &constant, const idDrawVert *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t xx=vdup_n_f32(constant[0]), yy=vdup_n_f32(constant[1]), zz=vdup_n_f32(constant[2]), ww=vdup_n_f32(constant[3]);
float tmp6[6];
for (_IX=0;_IX<_NM;_IX+=2) {
	memcpy(tmp6, &src[_IX].xyz, 3*sizeof(float));
	memcpy(tmp6+3, &src[_IX+1].xyz, 3*sizeof(float));
	float32x2x3_t src2=vld3_f32(tmp6);
	vst1_f32(dst+_IX, src2.val[0]*xx+src2.val[1]*yy+src2.val[2]*zz+ww);
}
if (_IX < count) dst[_IX] = constant.Normal() * src[_IX].xyz + constant[3];
}

/*
============
idSIMD_NEON::Dot

  dst[i] = src0[i] * src1[i];
============
*/
void VPCALL idSIMD_NEON::Dot(float *dst, const idVec3 *src0, const idVec3 *src1, const int count)
{
int _IX, _NM = count&0xfffffffe; 
for (_IX=0;_IX<_NM;_IX+=2) {
	float32x2x3_t src0_=vld3_f32((float32_t*)(src0+_IX));
	float32x2x3_t src1_=vld3_f32((float32_t*)(src1+_IX));
	vst1_f32(dst+_IX, src0_.val[0]*src1_.val[0]+src0_.val[1]*src1_.val[1]+src0_.val[2]*src1_.val[2]);
}
if (_IX < count) dst[_IX] = src0[_IX] * src1[_IX];
}

/*
============
idSIMD_NEON::CmpGT

  dst[i] = src0[i] > constant;
============
*/
void VPCALL idSIMD_NEON::CmpGT(byte *dst, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcgt_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] = tmp[0]?1:0; dst[_IX+1] = tmp[1]?1:0;
}
if (_IX < count) dst[_IX] = src0[_IX] > constant;
}

/*
============
idSIMD_NEON::CmpGT

  dst[i] |= ( src0[i] > constant ) << bitNum;
============
*/
void VPCALL idSIMD_NEON::CmpGT(byte *dst, const byte bitNum, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcgt_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] |= tmp[0]?1<< bitNum:0; dst[_IX+1] |= tmp[1]?1<< bitNum:0;
}
if (_IX < count) dst[_IX] = (src0[_IX] > constant) << bitNum;
}

/*
============
idSIMD_NEON::CmpGE

  dst[i] = src0[i] >= constant;
============
*/
void VPCALL idSIMD_NEON::CmpGE(byte *dst, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcge_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] = tmp[0]?1:0; dst[_IX+1] = tmp[1]?1:0;
}
if (_IX < count) dst[_IX] = src0[_IX] >= constant;
}

/*
============
idSIMD_NEON::CmpGE

  dst[i] |= ( src0[i] >= constant ) << bitNum;
============
*/
void VPCALL idSIMD_NEON::CmpGE(byte *dst, const byte bitNum, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcge_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] |= tmp[0]?1<< bitNum:0; dst[_IX+1] |= tmp[1]?1<< bitNum:0;
}
if (_IX < count) dst[_IX] = (src0[_IX] >= constant) << bitNum;
}

/*
============
idSIMD_NEON::CmpLT

  dst[i] = src0[i] < constant;
============
*/
void VPCALL idSIMD_NEON::CmpLT(byte *dst, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vclt_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] = tmp[0]?1:0; dst[_IX+1] = tmp[1]?1:0;
}
if (_IX < count) dst[_IX] = src0[_IX] < constant;
}

/*
============
idSIMD_NEON::CmpLT

  dst[i] |= ( src0[i] < constant ) << bitNum;
============
*/
void VPCALL idSIMD_NEON::CmpLT(byte *dst, const byte bitNum, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vclt_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] |= tmp[0]?1<< bitNum:0; dst[_IX+1] |= tmp[1]?1<< bitNum:0;
}
if (_IX < count) dst[_IX] = (src0[_IX] < constant) << bitNum;
}

/*
============
idSIMD_NEON::CmpLE

  dst[i] = src0[i] <= constant;
============
*/
void VPCALL idSIMD_NEON::CmpLE(byte *dst, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcle_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] = tmp[0]?1:0; dst[_IX+1] = tmp[1]?1:0;
}
if (_IX < count) dst[_IX] = src0[_IX] <= constant;
}

/*
============
idSIMD_NEON::CmpLE

  dst[i] |= ( src0[i] <= constant ) << bitNum;
============
*/
void VPCALL idSIMD_NEON::CmpLE(byte *dst, const byte bitNum, const float *src0, const float constant, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t constx2 = vdup_n_f32(constant);
for (_IX=0;_IX<_NM;_IX+=2) {
	uint32x2_t tmp = vcle_f32(vld1_f32(src0+_IX), constx2);
	dst[_IX] |= tmp[0]?1<< bitNum:0; dst[_IX+1] |= tmp[1]?1<< bitNum:0;
}
if (_IX < count) dst[_IX] = (src0[_IX] <= constant) << bitNum;
}


/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(float &min, float &max, const float *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t minx2 = vdup_n_f32(idMath::INFINITY);
float32x2_t maxx2 = vdup_n_f32(-idMath::INFINITY);
float32x2_t tmp;
for (_IX=0;_IX<_NM;_IX+=2) {
	tmp = vld1_f32(src+_IX);
	minx2 = vmin_f32(minx2, tmp);
	maxx2 = vmax_f32(maxx2, tmp);
}
if (_IX < count) {
	tmp = vdup_n_f32(src[_IX]);
	minx2 = vmin_f32(minx2, tmp);
	maxx2 = vmax_f32(maxx2, tmp);	
}
minx2 = vmin_f32(minx2, vdup_n_f32(minx2[1]));
maxx2 = vmax_f32(maxx2, vdup_n_f32(maxx2[1]));
min=minx2[0];
max=maxx2[0];
}

/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(idVec2 &min, idVec2 &max, const idVec2 *src, const int count)
{
float32x2_t minx2 = vdup_n_f32(idMath::INFINITY);
float32x2_t maxx2 = vdup_n_f32(-idMath::INFINITY);
float32x2_t tmp;
for(int _IX=0; _IX<count; _IX++) {
	tmp = vld1_f32((float*)(src+_IX));
	minx2 = vmin_f32(minx2, tmp);
	maxx2 = vmax_f32(maxx2, tmp);
}
vst1_f32((float*)&min, minx2);
vst1_f32((float*)&max, maxx2);
}

/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(idVec3 &min, idVec3 &max, const idVec3 *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2x3_t minx3 = {vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY)};
float32x2x3_t maxx3 = {vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY)};
float32x2x3_t tmp;
for (_IX=0;_IX<_NM;_IX+=2) {
	tmp = vld3_f32((float*)(src+_IX));
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
if (_IX < count) {
	tmp = {vdup_n_f32(src[_IX].x),  vdup_n_f32(src[_IX].y), vdup_n_f32(src[_IX].z)};
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
minx3.val[0] = vmin_f32(minx3.val[0], vdup_n_f32(minx3.val[0][1])); minx3.val[1] = vmin_f32(minx3.val[1], vdup_n_f32(minx3.val[1][1])); minx3.val[2] = vmin_f32(minx3.val[2], vdup_n_f32(minx3.val[2][1]));
maxx3.val[0] = vmax_f32(maxx3.val[0], vdup_n_f32(maxx3.val[0][1])); maxx3.val[1] = vmax_f32(maxx3.val[1], vdup_n_f32(maxx3.val[1][1])); maxx3.val[2] = vmax_f32(maxx3.val[2], vdup_n_f32(maxx3.val[2][1]));
min.x=minx3.val[0][0]; min.y=minx3.val[1][0]; min.z=minx3.val[2][0];
max.x=maxx3.val[0][0]; max.y=maxx3.val[1][0]; max.z=maxx3.val[2][0];
}

/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(idVec3 &min, idVec3 &max, const idDrawVert *src, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2x3_t minx3 = {vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY)};
float32x2x3_t maxx3 = {vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY)};
float tmp6[6];
float32x2x3_t tmp;
for (_IX=0;_IX<_NM;_IX+=2) {
	memcpy(tmp6, &src[_IX].xyz, 3*sizeof(float));
	memcpy(tmp6+3, &src[_IX+1].xyz, 3*sizeof(float));
	tmp = vld3_f32(tmp6);
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
if (_IX < count) {
	tmp = {vdup_n_f32(src[_IX].xyz.x),  vdup_n_f32(src[_IX].xyz.y), vdup_n_f32(src[_IX].xyz.z)};
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
minx3.val[0] = vmin_f32(minx3.val[0], vdup_n_f32(minx3.val[0][1])); minx3.val[1] = vmin_f32(minx3.val[1], vdup_n_f32(minx3.val[1][1])); minx3.val[2] = vmin_f32(minx3.val[2], vdup_n_f32(minx3.val[2][1]));
maxx3.val[0] = vmax_f32(maxx3.val[0], vdup_n_f32(maxx3.val[0][1])); maxx3.val[1] = vmax_f32(maxx3.val[1], vdup_n_f32(maxx3.val[1][1])); maxx3.val[2] = vmax_f32(maxx3.val[2], vdup_n_f32(maxx3.val[2][1]));
min.x=minx3.val[0][0]; min.y=minx3.val[1][0]; min.z=minx3.val[2][0];
max.x=maxx3.val[0][0]; max.y=maxx3.val[1][0]; max.z=maxx3.val[2][0];
}

/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(idVec3 &min, idVec3 &max, const idDrawVert *src, const int *indexes, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2x3_t minx3 = {vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY)};
float32x2x3_t maxx3 = {vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY)};
float tmp6[6];
float32x2x3_t tmp;
for (_IX=0;_IX<_NM;_IX+=2) {
	memcpy(tmp6, &src[indexes[_IX]].xyz, 3*sizeof(float));
	memcpy(tmp6+3, &src[indexes[_IX+1]].xyz, 3*sizeof(float));
	tmp = vld3_f32(tmp6);
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
if (_IX < count) {
	tmp = {vdup_n_f32(src[indexes[_IX]].xyz.x),  vdup_n_f32(src[indexes[_IX]].xyz.y), vdup_n_f32(src[indexes[_IX]].xyz.z)};
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
minx3.val[0] = vmin_f32(minx3.val[0], vdup_n_f32(minx3.val[0][1])); minx3.val[1] = vmin_f32(minx3.val[1], vdup_n_f32(minx3.val[1][1])); minx3.val[2] = vmin_f32(minx3.val[2], vdup_n_f32(minx3.val[2][1]));
maxx3.val[0] = vmax_f32(maxx3.val[0], vdup_n_f32(maxx3.val[0][1])); maxx3.val[1] = vmax_f32(maxx3.val[1], vdup_n_f32(maxx3.val[1][1])); maxx3.val[2] = vmax_f32(maxx3.val[2], vdup_n_f32(maxx3.val[2][1]));
min.x=minx3.val[0][0]; min.y=minx3.val[1][0]; min.z=minx3.val[2][0];
max.x=maxx3.val[0][0]; max.y=maxx3.val[1][0]; max.z=maxx3.val[2][0];
}

/*
============
idSIMD_NEON::MinMax
============
*/
void VPCALL idSIMD_NEON::MinMax(idVec3 &min, idVec3 &max, const idDrawVert *src, const short *indexes, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2x3_t minx3 = {vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY),vdup_n_f32(idMath::INFINITY)};
float32x2x3_t maxx3 = {vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY),vdup_n_f32(-idMath::INFINITY)};
float tmp6[6];
float32x2x3_t tmp;
for (_IX=0;_IX<_NM;_IX+=2) {
	memcpy(tmp6, &src[indexes[_IX]].xyz, 3*sizeof(float));
	memcpy(tmp6+3, &src[indexes[_IX+1]].xyz, 3*sizeof(float));
	tmp = vld3_f32(tmp6);
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
if (_IX < count) {
	tmp = {vdup_n_f32(src[indexes[_IX]].xyz.x),  vdup_n_f32(src[indexes[_IX]].xyz.y), vdup_n_f32(src[indexes[_IX]].xyz.z)};
	minx3.val[0] = vmin_f32(minx3.val[0], tmp.val[0]); minx3.val[1] = vmin_f32(minx3.val[1], tmp.val[1]); minx3.val[2] = vmin_f32(minx3.val[2], tmp.val[2]);
	maxx3.val[0] = vmax_f32(maxx3.val[0], tmp.val[0]); maxx3.val[1] = vmax_f32(maxx3.val[1], tmp.val[1]); maxx3.val[2] = vmax_f32(maxx3.val[2], tmp.val[2]);
}
minx3.val[0] = vmin_f32(minx3.val[0], vdup_n_f32(minx3.val[0][1])); minx3.val[1] = vmin_f32(minx3.val[1], vdup_n_f32(minx3.val[1][1])); minx3.val[2] = vmin_f32(minx3.val[2], vdup_n_f32(minx3.val[2][1]));
maxx3.val[0] = vmax_f32(maxx3.val[0], vdup_n_f32(maxx3.val[0][1])); maxx3.val[1] = vmax_f32(maxx3.val[1], vdup_n_f32(maxx3.val[1][1])); maxx3.val[2] = vmax_f32(maxx3.val[2], vdup_n_f32(maxx3.val[2][1]));
min.x=minx3.val[0][0]; min.y=minx3.val[1][0]; min.z=minx3.val[2][0];
max.x=maxx3.val[0][0]; max.y=maxx3.val[1][0]; max.z=maxx3.val[2][0];
}

/*
============
idSIMD_NEON::Clamp
============
*/
void VPCALL idSIMD_NEON::Clamp(float *dst, const float *src, const float min, const float max, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t minx2 = vdup_n_f32(min);
float32x2_t maxx2 = vdup_n_f32(max);
for (_IX=0;_IX<_NM;_IX+=2) {
	vst1_f32(dst+_IX, vmax_f32(minx2, vmin_f32(maxx2, vld1_f32(src+_IX))));
}
if (_IX < count) {
	dst[_IX] = src[_IX] < min ? min : src[_IX] > max ? max : src[_IX];
}
}

/*
============
idSIMD_NEON::ClampMin
============
*/
void VPCALL idSIMD_NEON::ClampMin(float *dst, const float *src, const float min, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t minx2 = vdup_n_f32(min);
for (_IX=0;_IX<_NM;_IX+=2) {
	vst1_f32(dst+_IX, vmax_f32(minx2, vld1_f32(src+_IX)));
}
if (_IX < count) {
	dst[_IX] = src[_IX] < min ? min : src[_IX];
}
}

/*
============
idSIMD_NEON::ClampMax
============
*/
void VPCALL idSIMD_NEON::ClampMax(float *dst, const float *src, const float max, const int count)
{
int _IX, _NM = count&0xfffffffe; 
float32x2_t maxx2 = vdup_n_f32(max);
for (_IX=0;_IX<_NM;_IX+=2) {
	vst1_f32(dst+_IX, vmin_f32(maxx2, vld1_f32(src+_IX)));
}
if (_IX < count) {
	dst[_IX] = src[_IX] > max ? max : src[_IX];
}
}

/*
============
idSIMD_NEON::Copy16
============
*/
void VPCALL idSIMD_NEON::Copy16(float *dst, const float *src, const int count)
{
	memcpy(dst, src, count*sizeof(count));
}

/*
============
idSIMD_NEON::Add16
============
*/
void VPCALL idSIMD_NEON::Add16(float *dst, const float *src1, const float *src2, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vaddq_f32(vld1q_f32(src1+i), vld1q_f32(src2+i)));
}

/*
============
idSIMD_NEON::Sub16
============
*/
void VPCALL idSIMD_NEON::Sub16(float *dst, const float *src1, const float *src2, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vsubq_f32(vld1q_f32(src1+i), vld1q_f32(src2+i)));
}

/*
============
idSIMD_NEON::Mul16
============
*/
void VPCALL idSIMD_NEON::Mul16(float *dst, const float *src1, const float constant, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vmulq_n_f32(vld1q_f32(src1+i), constant));
}

/*
============
idSIMD_NEON::AddAssign16
============
*/
void VPCALL idSIMD_NEON::AddAssign16(float *dst, const float *src, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vaddq_f32(vld1q_f32(dst+i), vld1q_f32(src+i)));
}

/*
============
idSIMD_NEON::SubAssign16
============
*/
void VPCALL idSIMD_NEON::SubAssign16(float *dst, const float *src, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vsubq_f32(vld1q_f32(dst+i), vld1q_f32(src+i)));
}

/*
============
idSIMD_NEON::MulAssign16
============
*/
void VPCALL idSIMD_NEON::MulAssign16(float *dst, const float constant, const int count)
{
	int count2 = (count + 3) & ~3;
	for (int i = 0; i<count2; i+=4)
		vst1q_f32(dst+i, vmulq_n_f32(vld1q_f32(dst+i), constant));
}

/*
============
idSIMD_NEON::BlendJoints
============
*/

void VPCALL idSIMD_NEON::BlendJoints(idJointQuat *joints, const idJointQuat *blendJoints, const float lerp, const int *index, const int numJoints)
{
	int i;

	if (lerp <= 0.0f) {
		return;
	} else if (lerp >= 1.0f) {
		for (i = 0; i < numJoints; i++) {
			int j = index[i];
			joints[j] = blendJoints[j];
		}

		return;
	}

	for (i = 0; i <= numJoints - 4; i += 4) {
		float32x4_t jointVert0;
		float32x4_t jointVert1;
		float32x4_t jointVert2;
		float32x4_t blendVert0;
		float32x4_t blendVert1;
		float32x4_t blendVert2;
		float32x4_t jointQuat0;
		float32x4_t jointQuat1;
		float32x4_t jointQuat2;
		float32x4_t jointQuat3;
		float32x4_t blendQuat0;
		float32x4_t blendQuat1;
		float32x4_t blendQuat2;
		float32x4_t blendQuat3;

		for (int j = 0; j < 4; j++) {
			int n = index[i+j];

			jointVert0[j] = joints[n].t[0];
			jointVert1[j] = joints[n].t[1];
			jointVert2[j] = joints[n].t[2];

			blendVert0[j] = blendJoints[n].t[0];
			blendVert1[j] = blendJoints[n].t[1];
			blendVert2[j] = blendJoints[n].t[2];

			jointQuat0[j] = joints[n].q[0];
			jointQuat1[j] = joints[n].q[1];
			jointQuat2[j] = joints[n].q[2];
			jointQuat3[j] = joints[n].q[3];

			blendQuat0[j] = blendJoints[n].q[0];
			blendQuat1[j] = blendJoints[n].q[1];
			blendQuat2[j] = blendJoints[n].q[2];
			blendQuat3[j] = blendJoints[n].q[3];
		}

		jointVert0 += lerp * (blendVert0 - jointVert0);

		jointVert1 += lerp * (blendVert1 - jointVert1);

		jointVert2 += lerp * (blendVert2 - jointVert2);

		float32x4_t cosom;
		float32x4_t sinom;
		float32x4_t omega0;
		float32x4_t omega1;
		float32x4_t scale0;
		float32x4_t scale1;
		uint32x4_t signBit;

		cosom = jointQuat0 * blendQuat0;
		cosom += jointQuat1 * blendQuat1;
		cosom += jointQuat2 * blendQuat2;
		cosom += jointQuat3 * blendQuat3;

		signBit = vandq_u32(vreinterpretq_u32_f32(cosom), vdupq_n_u32(1<<31));
		cosom = vreinterpretq_f32_u32(veorq_u32(vreinterpretq_u32_f32(cosom), signBit));

		scale0 = 1.0f - cosom * cosom;

		uint32x4_t infZero;
		infZero = vcleq_f32(scale0, vdupq_n_f32(0.0f));
		scale0 = vreinterpretq_f32_u32(vorrq_u32(vandq_u32(vreinterpretq_u32_f32(SIMD_SP_tiny), infZero), vandq_u32(vreinterpretq_u32_f32(scale0), vmvnq_u32(infZero))));

		sinom[0] = idMath::RSqrt(scale0[0]);
		sinom[1] = idMath::RSqrt(scale0[1]);
		sinom[2] = idMath::RSqrt(scale0[2]);
		sinom[3] = idMath::RSqrt(scale0[3]);

		scale0 *= sinom;

		omega0[0] = idMath::ATan16(scale0[0], cosom[0]);
		omega0[1] = idMath::ATan16(scale0[1], cosom[1]);
		omega0[2] = idMath::ATan16(scale0[2], cosom[2]);
		omega0[3] = idMath::ATan16(scale0[3], cosom[3]);

		omega1 = lerp * omega0;

		omega0 -= omega1;

		scale0[0] = idMath::Sin16(omega0[0]) * sinom[0];
		scale0[1] = idMath::Sin16(omega0[1]) * sinom[1];
		scale0[2] = idMath::Sin16(omega0[2]) * sinom[2];
		scale0[3] = idMath::Sin16(omega0[3]) * sinom[3];

		scale1[0] = idMath::Sin16(omega1[0]) * sinom[0];
		scale1[1] = idMath::Sin16(omega1[1]) * sinom[1];
		scale1[2] = idMath::Sin16(omega1[2]) * sinom[2];
		scale1[3] = idMath::Sin16(omega1[3]) * sinom[3];

		scale1 = vreinterpretq_f32_u32(veorq_u32(vreinterpretq_u32_f32(scale1), signBit));

		jointQuat0 = scale0 * jointQuat0 + scale1 * blendQuat0;
		jointQuat1 = scale0 * jointQuat1 + scale1 * blendQuat1;
		jointQuat2 = scale0 * jointQuat2 + scale1 * blendQuat2;
		jointQuat3 = scale0 * jointQuat3 + scale1 * blendQuat3;

		for (int j = 0; j < 4; j++) {
			int n = index[i+j];

			joints[n].t[0] = jointVert0[j];
			joints[n].t[1] = jointVert1[j];
			joints[n].t[2] = jointVert2[j];

			joints[n].q[0] = jointQuat0[j];
			joints[n].q[1] = jointQuat1[j];
			joints[n].q[2] = jointQuat2[j];
			joints[n].q[3] = jointQuat3[j];
		}
	}

	for (; i < numJoints; i++) {
		int n = index[i];

		idVec3 &jointVert = joints[n].t;
		const idVec3 &blendVert = blendJoints[n].t;

		jointVert[0] += lerp * (blendVert[0] - jointVert[0]);
		jointVert[1] += lerp * (blendVert[1] - jointVert[1]);
		jointVert[2] += lerp * (blendVert[2] - jointVert[2]);

		idQuat &jointQuat = joints[n].q;
		const idQuat &blendQuat = blendJoints[n].q;

		float cosom;
		float sinom;
		float omega;
		float scale0;
		float scale1;
		unsigned long signBit;

		cosom = jointQuat.x * blendQuat.x + jointQuat.y * blendQuat.y + jointQuat.z * blendQuat.z + jointQuat.w * blendQuat.w;

		signBit = (*(unsigned long *)&cosom) & (1 << 31);

		(*(unsigned long *)&cosom) ^= signBit;

		scale0 = 1.0f - cosom * cosom;
		scale0 = (scale0 <= 0.0f) ? SIMD_SP_tiny[0] : scale0;
		sinom = idMath::InvSqrt(scale0);
		omega = idMath::ATan16(scale0 * sinom, cosom);
		scale0 = idMath::Sin16((1.0f - lerp) * omega) * sinom;
		scale1 = idMath::Sin16(lerp * omega) * sinom;

		(*(unsigned long *)&scale1) ^= signBit;

		jointQuat.x = scale0 * jointQuat.x + scale1 * blendQuat.x;
		jointQuat.y = scale0 * jointQuat.y + scale1 * blendQuat.y;
		jointQuat.z = scale0 * jointQuat.z + scale1 * blendQuat.z;
		jointQuat.w = scale0 * jointQuat.w + scale1 * blendQuat.w;
	}
}

/*
============
idSIMD_NEON::ConvertJointQuatsToJointMats
============
*/
void VPCALL idSIMD_NEON::ConvertJointQuatsToJointMats(idJointMat *jointMats, const idJointQuat *jointQuats, const int numJoints)
{

	assert(sizeof(idJointQuat) == JOINTQUAT_SIZE);
	assert(sizeof(idJointMat) == JOINTMAT_SIZE);
	assert((int)(&((idJointQuat *)0)->t) == (int)(&((idJointQuat *)0)->q) + (int)sizeof(((idJointQuat *)0)->q));

	for (int i = 0; i < numJoints; i++) {

		const float *q = jointQuats[i].q.ToFloatPtr();
		float *m = jointMats[i].ToFloatPtr();

		m[0*4+3] = q[4];
		m[1*4+3] = q[5];
		m[2*4+3] = q[6];

		float x2 = q[0] + q[0];
		float y2 = q[1] + q[1];
		float z2 = q[2] + q[2];

		{
			float xx = q[0] * x2;
			float yy = q[1] * y2;
			float zz = q[2] * z2;

			m[0*4+0] = 1.0f - yy - zz;
			m[1*4+1] = 1.0f - xx - zz;
			m[2*4+2] = 1.0f - xx - yy;
		}

		{
			float yz = q[1] * z2;
			float wx = q[3] * x2;

			m[2*4+1] = yz - wx;
			m[1*4+2] = yz + wx;
		}

		{
			float xy = q[0] * y2;
			float wz = q[3] * z2;

			m[1*4+0] = xy - wz;
			m[0*4+1] = xy + wz;
		}

		{
			float xz = q[0] * z2;
			float wy = q[3] * y2;

			m[0*4+2] = xz - wy;
			m[2*4+0] = xz + wy;
		}
	}
}

/*
============
idSIMD_NEON::ConvertJointMatsToJointQuats
============
*/
void VPCALL idSIMD_NEON::ConvertJointMatsToJointQuats(idJointQuat *jointQuats, const idJointMat *jointMats, const int numJoints)
{

	assert(sizeof(idJointQuat) == JOINTQUAT_SIZE);
	assert(sizeof(idJointMat) == JOINTMAT_SIZE);
	assert((int)(&((idJointQuat *)0)->t) == (int)(&((idJointQuat *)0)->q) + (int)sizeof(((idJointQuat *)0)->q));


	for (int i = 0; i < numJoints; i++) {

		float *q = jointQuats[i].q.ToFloatPtr();
		const float *m = jointMats[i].ToFloatPtr();

		if (m[0 * 4 + 0] + m[1 * 4 + 1] + m[2 * 4 + 2] > 0.0f) {

			float t = + m[0 * 4 + 0] + m[1 * 4 + 1] + m[2 * 4 + 2] + 1.0f;
			float s = idMath::InvSqrt(t) * 0.5f;

			q[3] = s * t;
			q[2] = (m[0 * 4 + 1] - m[1 * 4 + 0]) * s;
			q[1] = (m[2 * 4 + 0] - m[0 * 4 + 2]) * s;
			q[0] = (m[1 * 4 + 2] - m[2 * 4 + 1]) * s;

		} else if (m[0 * 4 + 0] > m[1 * 4 + 1] && m[0 * 4 + 0] > m[2 * 4 + 2]) {

			float t = + m[0 * 4 + 0] - m[1 * 4 + 1] - m[2 * 4 + 2] + 1.0f;
			float s = idMath::InvSqrt(t) * 0.5f;

			q[0] = s * t;
			q[1] = (m[0 * 4 + 1] + m[1 * 4 + 0]) * s;
			q[2] = (m[2 * 4 + 0] + m[0 * 4 + 2]) * s;
			q[3] = (m[1 * 4 + 2] - m[2 * 4 + 1]) * s;

		} else if (m[1 * 4 + 1] > m[2 * 4 + 2]) {

			float t = - m[0 * 4 + 0] + m[1 * 4 + 1] - m[2 * 4 + 2] + 1.0f;
			float s = idMath::InvSqrt(t) * 0.5f;

			q[1] = s * t;
			q[0] = (m[0 * 4 + 1] + m[1 * 4 + 0]) * s;
			q[3] = (m[2 * 4 + 0] - m[0 * 4 + 2]) * s;
			q[2] = (m[1 * 4 + 2] + m[2 * 4 + 1]) * s;

		} else {

			float t = - m[0 * 4 + 0] - m[1 * 4 + 1] + m[2 * 4 + 2] + 1.0f;
			float s = idMath::InvSqrt(t) * 0.5f;

			q[2] = s * t;
			q[3] = (m[0 * 4 + 1] - m[1 * 4 + 0]) * s;
			q[0] = (m[2 * 4 + 0] + m[0 * 4 + 2]) * s;
			q[1] = (m[1 * 4 + 2] + m[2 * 4 + 1]) * s;

		}

		q[4] = m[0 * 4 + 3];
		q[5] = m[1 * 4 + 3];
		q[6] = m[2 * 4 + 3];
	}

}


/*
============
idSIMD_NEON::CreateShadowCache
============
*/
int VPCALL idSIMD_NEON::CreateShadowCache(idVec4 *vertexCache, int *vertRemap, const idVec3 &lightOrigin, const idDrawVert *verts, const int numVerts)
{
	int outVerts = 0;

	register float32x4_t l = vld1q_f32((float*)&lightOrigin);
	l[3] = 1.0f;

	for (int i = 0; i < numVerts; i++) {
		if (vertRemap[i]) {
			continue;
		}

		register float32x4_t v=vld1q_f32(verts[i].xyz.ToFloatPtr());
		v[3] = l[3];

		vst1q_f32((float*)&vertexCache[outVerts+0], v);

		// R_SetupProjection() builds the projection matrix with a slight crunch
		// for depth, which keeps this w=0 division from rasterizing right at the
		// wrap around point and causing depth fighting with the rear caps
		vst1q_f32((float*)&vertexCache[outVerts+1], vsubq_f32(v, l));

		vertRemap[i] = outVerts;

		outVerts += 2;
	}

	return outVerts;
}

/*
============
idSIMD_NEON::CreateVertexProgramShadowCache
============
*/
/*int VPCALL idSIMD_NEON::CreateVertexProgramShadowCache(idVec4 *vertexCache, const idDrawVert *verts, const int numVerts)
{
	for (int i = 0; i < numVerts; i++) {
		const float *v = verts[i].xyz.ToFloatPtr();
		memcpy(&vertexCache[i*2+0], v, 3*sizeof(float));
		memcpy(&vertexCache[i*2+1], v, 3*sizeof(float));
		vertexCache[i*2+0][3] = 1.0f;
		vertexCache[i*2+1][3] = 0.0f;
	}

	return numVerts * 2;
}*/ //Less efficiant then generic


/*
============
idSIMD_NEON::UpSamplePCMTo44kHz

  Duplicate samples for 44kHz output.
============
*/
void idSIMD_NEON::UpSamplePCMTo44kHz(float *dest, const short *src, const int numSamples, const int kHz, const int numChannels)
{
	int i = 0;
	if (kHz == 11025) {
		if (numChannels == 1) {
			for (; i < numSamples; i++) {
				vst1q_f32(dest+i*4, vcvtq_f32_s32(vmovl_s16(vdup_n_s16(src[i+0]))));
			}
		} else {
			for (int i = 0; i < numSamples; i += 2) {
				dest[i*4+0] = dest[i*4+2] = dest[i*4+4] = dest[i*4+6] = (float) src[i+0];
				dest[i*4+1] = dest[i*4+3] = dest[i*4+5] = dest[i*4+7] = (float) src[i+1];
			}
		}
	} else if (kHz == 22050) {
		if (numChannels == 1) {
			for (int i = 0; i < numSamples; i++) {
				vst1_f32(dest+i*2, vdup_n_f32((float)src[i+0]));
			}
		} else {
			for (int i = 0; i < numSamples; i += 2) {
				dest[i*2+0] = dest[i*2+2] = (float) src[i+0];
				dest[i*2+1] = dest[i*2+3] = (float) src[i+1];
			}
		}
	} else if (kHz == 44100) {
		for (;  NOT_4BYTE_ALIGNED(src[i]) && (i < numSamples); i++)
			dest[i] = (float) src[i];
		for (; i+3 < numSamples; i+=4)
			vst1q_f32(dest+i, vcvtq_f32_s32(vmovl_s16(vld1_s16(src+i))));
		for (;i < numSamples; i++)
			dest[i] = (float) src[i];
	} else {
		assert(0);
	}
}


/*
============
idSIMD_NEON::UpSampleOGGTo44kHz

  Duplicate samples for 44kHz output.
============
*/
void idSIMD_NEON::UpSampleOGGTo44kHz(float *dest, const float *const *ogg, const int numSamples, const int kHz, const int numChannels)
{
	int i = 0;
	if (kHz == 11025) {
		if (numChannels == 1) {
			for (; i < numSamples; i++) {
				vst1q_f32(dest+i*4, vmulq_n_f32(vdupq_n_f32(ogg[0][i]), 32768.0f));
			}
		} else {
			for (; i < numSamples >> 1; i++) {
				float32x2_t tmp = {ogg[0][i], ogg[1][i]};
				tmp *= 32768.0f;
				vst1_f32(dest+i*8, tmp); vst1_f32(dest+i*8+2, tmp); vst1_f32(dest+i*8+4, tmp); vst1_f32(dest+i*8+6, tmp);
			}
		}
	} else if (kHz == 22050) {
		if (numChannels == 1) {
			for (; i < numSamples; i++) {
				vst1_f32(dest+i*2, vmul_n_f32(vdup_n_f32(ogg[0][i]), 32768.0f));
			}
		} else {
			for (; i < numSamples >> 1; i++) {
				float32x2_t tmp = {ogg[0][i], ogg[1][i]};
				tmp *= 32768.0f;
				vst1_f32(dest+i*4, tmp); vst1_f32(dest+i*4+2, tmp);
			}
		}
	} else if (kHz == 44100) {
		if (numChannels == 1) {
			for (; i+1 < numSamples; i+=2) {
				vst1_f32(dest+i, vmul_n_f32(vld1_f32(&ogg[0][i]), 32768.0f));
			if(i<numSamples)
				dest[i*1+0] = ogg[0][i] * 32768.0f;
			}
		} else {
			for (int i = 0; i+1 < numSamples >> 1; i+=2) {
				float32x2_t tmp1,tmp2;
				tmp1 = vld1_f32(&ogg[0][i])*32768.0f; tmp2 = vld1_f32(&ogg[1][i])*32768.0f;
				float32x2x2_t tmp = vtrn_f32(tmp1, tmp2);
				vst1_f32(dest+i*2+0, tmp.val[0]);
				vst1_f32(dest+i*2+2, tmp.val[1]);
			}
			if (i < numSamples >> 1) {
				dest[i*2+0] = ogg[0][i] * 32768.0f;
				dest[i*2+1] = ogg[1][i] * 32768.0f;
			}

		}
	} else {
		assert(0);
	}
}

/*
============
idSIMD_NEON::MixSoundTwoSpeakerMono
============
*/
/*void VPCALL idSIMD_NEON::MixSoundTwoSpeakerMono(float *mixBuffer, const float *samples, const int numSamples, const float lastV[2], const float currentV[2])
{
	float32x2_t sLR = vld1_f32(lastV);
	const float32x2_t incLR = vmul_n_f32(vsub_f32(vld1_f32(currentV), sLR), 1.0/MIXBUFFER_SAMPLES);

	for (int j = 0; j < MIXBUFFER_SAMPLES; j++) {
		vst1_f32(mixBuffer+j*2, vmla_f32(vld1_f32(mixBuffer+j*2), vdup_n_f32(samples[j]), sLR));
		sLR = vadd_f32(sLR, incLR);
	}
}*/ //Less efficiant then generic

/*
============
idSIMD_NEON::MixSoundTwoSpeakerStereo
============
*/
/*void VPCALL idSIMD_NEON::MixSoundTwoSpeakerStereo(float *mixBuffer, const float *samples, const int numSamples, const float lastV[2], const float currentV[2])
{
	float32x2_t sLR = vld1_f32(lastV);
	const float32x2_t incLR = vmul_n_f32(vsub_f32(vld1_f32(currentV), sLR), 1.0/MIXBUFFER_SAMPLES);

	for (int j = 0; j < MIXBUFFER_SAMPLES; j++) {
		vst1_f32(mixBuffer+j*2, vmla_f32(vld1_f32(mixBuffer+j*2), vld1_f32(samples+j*2), sLR));
		sLR = vadd_f32(sLR, incLR);
	}
}*/ // Less efficiant than generic

/*
============
idSIMD_NEON::MixedSoundToSamples
============
*/
void VPCALL idSIMD_NEON::MixedSoundToSamples(short *samples, const float *mixBuffer, const int numSamples)
{
	int i=0;
	//float32x4_t valmin=vdupq_n_f32(-32768.0f), valmax=vdupq_n_f32(32767.0f);
	for (;  NOT_4BYTE_ALIGNED(samples[i]) && (i < numSamples); i++) {
		samples[i] = mixBuffer[i] <= -32768.0f ? -32768 : mixBuffer[i] >= 32767.0f ? 32767 : (short) mixBuffer[i];
	}
	register float32x4_t sampx4;
	for (; i+3 < numSamples; i+=4) {
		sampx4=vld1q_f32(mixBuffer+i);
		//sampx4=vminq_f32(valmin, vmaxq_f32(valmax, sampx4));
		vst1_s16(samples+i, vqmovn_s32(vcvtq_s32_f32(sampx4)));
	}
	for (; i < numSamples; i++) {
		samples[i] = mixBuffer[i] <= -32768.0f ? -32768 : mixBuffer[i] >= 32767.0f ? 32767 : (short) mixBuffer[i];
	}
}

#endif	//__ARM_NEON__